//
// Dmitry Brant
// Assignment 5, CIS 335, Prof. Jackie Woldering
// EXTRA CREDIT VERSION
//
//-------------------------------------
#include <vcl.h>
#pragma hdrstop

#include <stdio.h>
#include <stdlib.h>
#include <stdexcept>
//---------------------------------------------------------------------------

using namespace std;

int currentPass=0;
int PROGADDR, EXECADDR, linkPtr;
TList* ESTAB;
TList* Progs;
//----------------------------------

typedef struct{
    char name[7];
    int number;
}SIC_Ref;

//Only two functions in this one
void Pass1(char* fileName);
void Pass2();

//Class that encapsulates a SIC/XE control section
class ControlSection{
private:
    unsigned char* buffer;
    unsigned char* bufferMask;
    int bufPtr, startAddr, progSize;
    TList *importList;
    TStringList* Modifications;
    String progName;
public:
    ControlSection(int _startAddr, char* startString):
      bufPtr(0), startAddr(_startAddr){
        buffer = new unsigned char[1024]; //should be enough, eh?
        bufferMask = new unsigned char[1024];
        memset(buffer, 0, 1024);
        memset(bufferMask, 1, 1024);
        importList = new TList;
        Modifications = new TStringList;
        int ptr;
        String tmpStr;
        for(ptr=1; ptr<7; ptr++) tmpStr += String(startString[ptr]);
        progName = tmpStr;
        tmpStr = "0x";
        for(; ptr<13; ptr++) tmpStr += String(startString[ptr]);
        tmpStr = "0x";
        for(; ptr<19; ptr++) tmpStr += String(startString[ptr]);
        sscanf(tmpStr.c_str(), "%x", &progSize);
        //add ourselves to externals
        SIC_Ref* symbol = new SIC_Ref;
        strcpy(symbol->name, progName.c_str());
        symbol->number = startAddr;
        ESTAB->Add(symbol);
    }
    ~ControlSection(){
        delete [] buffer;
        delete [] bufferMask;
        for(int i=0; i<importList->Count; i++){ delete ((SIC_Ref*)importList->Items[i]); }
        delete importList;
        delete Modifications;
    }

    int GetProgSize(){ return(progSize); }
    unsigned char* GetBuffer(){ return(buffer); }
    unsigned char* GetBufferMask(){ return(bufferMask); }
    String GetName(){ return(progName); }
    int GetStartAddr(){ return(startAddr); }

    void AddExports(char* exports){
        if(exports[0] != 'D') return;
        int ptr=1, exportAddress;
        String tmpStr;
        while(exports[ptr] != 0){
            tmpStr = "";
            for(int i=0; i<6; i++) tmpStr += String(exports[ptr++]);
            //check if this is a duplicate
            for(int i=0; i<ESTAB->Count; i++){
                if(!strcmp(((SIC_Ref*)ESTAB->Items[i])->name, tmpStr.c_str())){
                    throw runtime_error(("Duplicate External Symbol " + tmpStr).c_str());
                }
            }
            SIC_Ref* s = new SIC_Ref;
            ESTAB->Add(s);
            strcpy(s->name, tmpStr.c_str());
            tmpStr = "0x";
            for(int i=0; i<6; i++) tmpStr += String(exports[ptr++]);
            sscanf(tmpStr.c_str(), "%x", &exportAddress);
            s->number = exportAddress + startAddr;
        }
    }

    void AddImports(char* imports){
        if(imports[0] != 'R') return;
        int ptr=1, importNum;
        String tmpStr;
        while(imports[ptr] != 0){
            tmpStr = "0x";
            for(int i=0; i<2; i++) tmpStr += String(imports[ptr++]);
            sscanf(tmpStr.c_str(), "%x", &importNum);
            tmpStr = "";
            for(int i=0; i<6; i++){
                tmpStr += String(imports[ptr++]);
                if(imports[ptr] == 0){ tmpStr += AnsiString::StringOfChar(' ', 6-tmpStr.Length()); break; }
            }
            SIC_Ref* s = new SIC_Ref;
            importList->Add(s);
            strcpy(s->name, tmpStr.c_str());
            s->number = importNum;
        }
        //must add obligatory import of self
        SIC_Ref* s = new SIC_Ref;
        importList->Add(s);
        strcpy(s->name, progName.c_str());
        s->number = 1;
    }

    void AddTextRecord(char* textrecord){
        if(textrecord[0] != 'T') return;
        int ptr, lineStartAddr, lineSize;
        //obtain starting address
        String tmpStr="0x";
        for(ptr=1; ptr<7; ptr++) tmpStr += String(textrecord[ptr]);
        sscanf(tmpStr.c_str(), "%x", &lineStartAddr);
        tmpStr = "0x";
        for(; ptr<9; ptr++) tmpStr += String(textrecord[ptr]);
        sscanf(tmpStr.c_str(), "%x", &lineSize);
        bufPtr = lineStartAddr;
        for(int i=0; i<lineSize; i++){
            int thisByte;
            tmpStr = "0x";
            tmpStr += String(textrecord[ptr++]);
            tmpStr += String(textrecord[ptr++]);
            sscanf(tmpStr.c_str(), "%x", &thisByte);
            bufferMask[bufPtr] = 2;
            buffer[bufPtr++] = thisByte;
        }
    }

    void AddModification(char* modification){
        Modifications->Add(String(modification));
    }

    void ApplyModifications(){
        for(int i=0; i<Modifications->Count; i++){
            char* modification = Modifications->Strings[i].c_str();
            int ptr, modAddr, modNumHalfBytes, modSymbolNum, operand, leftover;
            bool modPlus=false;
            String tmpStr="0x";
            for(ptr=1; ptr<7; ptr++) tmpStr += String(modification[ptr]);
            sscanf(tmpStr.c_str(), "%x", &modAddr);
            tmpStr = "0x";
            for(; ptr<9; ptr++) tmpStr += String(modification[ptr]);
            sscanf(tmpStr.c_str(), "%x", &modNumHalfBytes);
            tmpStr = "0x";
            if(modification[ptr++] == '+') modPlus = true; else modPlus = false;
            for(; ptr<12; ptr++) tmpStr += String(modification[ptr]);
            sscanf(tmpStr.c_str(), "%x", &modSymbolNum);
            //get the address as it is now
            tmpStr = "0x";
            int numBytes = (modNumHalfBytes+1)/2;
            for(int i=modAddr; i<(modAddr+numBytes); i++)
                tmpStr += IntToHex((int)buffer[i], 2);
            sscanf(tmpStr.c_str(), "%x", &operand);
            if(modNumHalfBytes%2){ leftover = ((operand >> (4*modNumHalfBytes)) & 0xF); }
            //operand += startAddr;
            //find the symbol in the table
            SIC_Ref* symbol=NULL;
            for(int i=0; i<importList->Count; i++){
                if(((SIC_Ref*)importList->Items[i])->number == modSymbolNum) symbol = (SIC_Ref*)importList->Items[i];
            }
            //search for it in the externals
            SIC_Ref* external=NULL;
            for(int i=0; i<ESTAB->Count; i++){
                if(!strcmp(((SIC_Ref*)ESTAB->Items[i])->name, symbol->name)) external = (SIC_Ref*)ESTAB->Items[i];
            }
            if(!external)
                throw runtime_error(("Unresolved External Symbol " + String(symbol->name)).c_str());
            if(modPlus)
                operand += external->number;
            else
                operand -= external->number;
            //now put it back
            if(modNumHalfBytes%2){
                operand &= 0xFFFFF;
                operand |= (leftover << (4*modNumHalfBytes));
            }
            for(int i=0; i<numBytes; i++){
                buffer[modAddr+numBytes-i-1] = (operand & 0xFF);
                operand >>= 8;
            }
        }
    }
};   //class
//------------------------------------------------


int main(int argc, char* argv[]){
     //identify ourselves
    printf("SIC/XE Linker/Loader Version 1.0\n");
    printf("Copyright Dmitry Brant, Cleveland State University, June 2003\n\n");
    //check command line for input file(s)
    if(argc < 2){
        printf("Usage: siclink inputfile [inputfile [...]] [-a <address>]\n");
        return(0);
    }
    String fileName;
    if(argc > 1) fileName = String(argv[1]);
    TStringList* theFiles = new TStringList();

    bool gotAddress=false;
    for(int i=1; i<argc; i++){
        String tempStr = String(argv[i]);
        if((tempStr == "-a") || (tempStr == "-A")){
            if((i+1) < argc){
                sscanf(argv[++i], "%x", &PROGADDR);
                gotAddress = true;
            }
        }
        else{
            theFiles->Add(String(argv[i]));
        }
    }
    if(!gotAddress){
        printf("Starting address not given. Please Specify (in hex): ");
        scanf("%x", &PROGADDR);
        gotAddress = true;
    }
    printf("Using starting address: %x\n", PROGADDR);
    ESTAB = new TList;
    Progs = new TList;

    try{

        //here we go
        linkPtr = PROGADDR;
        for(int i=0; i<theFiles->Count; i++)
            Pass1(theFiles->Strings[i].c_str());
        Pass2();
        printf("Done.\n");

    }catch(exception &e){
        printf("Error: %s", e.what());
        printf(" (Pass %d)\n", currentPass);
    }

    delete theFiles;
    for(int i=0; i<Progs->Count; i++) delete ((ControlSection*)Progs->Items[i]);
    delete Progs;
    for(int i=0; i<ESTAB->Count; i++) delete ((SIC_Ref*)ESTAB->Items[i]);
    delete ESTAB;
    return(0);
}
//---------------------------------------------------------------------------


void Pass1(char* fileName){
    currentPass = 1;
    FILE* f = fopen(fileName, "r");
    fseek(f, 0L, SEEK_END);
    int fileLen = ftell(f);
    fseek(f, 0, SEEK_SET);
    char* buffer = new char[fileLen+1];
    try{
        if(!fread(buffer, 1, fileLen, f)){
            fclose(f);
            throw runtime_error("Could not read from specified file.");
        }
        buffer[fileLen] = 0;
        fclose(f);

        //Tokenize the sucker
        char* token = strtok(buffer, "\r\n");
        ControlSection* currentProg;
        while(token){
            if(token[0] == 'H'){
                //new header record -- new program!
                ControlSection* sec = new ControlSection(linkPtr, token);
                Progs->Add(sec);
                currentProg = sec;
            }
            else if(token[0] == 'D')
                currentProg->AddExports(token);
            else if(token[0] == 'R')
                currentProg->AddImports(token);
            else if(token[0] == 'T')
                currentProg->AddTextRecord(token);
            else if(token[0] == 'M')
                currentProg->AddModification(token);
            else if(token[0] == 'E'){
                //is there an exec address?
                if(token[1]){
                    int addr;
                    String tmpStr="0x";
                    for(int ptr=1; ptr<7; ptr++) tmpStr += String(token[ptr]);
                    sscanf(tmpStr.c_str(), "%x", &addr);
                    EXECADDR = addr + linkPtr;
                }
                //update link pointer
                linkPtr += currentProg->GetProgSize();
            }
            else
                throw runtime_error("Unknown Record Type");

            token = strtok(NULL, "\r\n");
        }
    }__finally{
        delete [] buffer;
    }
}
//---------------------------------------------------------------------------


void Pass2(){
    String output;
    currentPass = 2;

    //Apply all modifications
    for(int i=0; i<Progs->Count; i++)
        ((ControlSection*)Progs->Items[i])->ApplyModifications();

    unsigned char* MEM = new unsigned char[32768];
    unsigned char* memMask = new unsigned char[32768];
    //memset it to zero for better clarity
    memset(MEM, 0, 32768);
    memset(memMask, 0, 32768);
    try{
        //graft each program section onto memory array
        int memPtr = PROGADDR;
        for(int i=0; i<Progs->Count; i++){
            ControlSection* prog = (ControlSection*)Progs->Items[i];
            memcpy(&MEM[memPtr], prog->GetBuffer(), prog->GetProgSize());
            memcpy(&memMask[memPtr], prog->GetBufferMask(), prog->GetProgSize());
            memPtr += prog->GetProgSize();
        }

        //print symbols
        output += "Section      Symbol      Address     Length\r\n-------------------------------------------\r\n";
        for(int i=0; i<Progs->Count; i++){
            ControlSection* sect = (ControlSection*)Progs->Items[i];
            output += sect->GetName() + "                   ";
            output += IntToHex((int)sect->GetStartAddr(), 4) + "        ";
            output += IntToHex((int)sect->GetProgSize(), 4) + "\r\n";
            unsigned int nextAddrBound;
            if(i<(Progs->Count-1)) nextAddrBound = ((ControlSection*)Progs->Items[i+1])->GetStartAddr();
            else nextAddrBound = 0xFFFFFFFF;
            for(int i=0; i<ESTAB->Count; i++){
                SIC_Ref* symbol = (SIC_Ref*)ESTAB->Items[i];
                if((symbol->number <= sect->GetStartAddr()) ||
                   (symbol->number >= nextAddrBound)) continue;
                output += "             " + String(symbol->name);
                output += "      ";
                output += IntToHex((int)symbol->number, 4);
                output += "\r\n";
            }
        }
        output += "-------------------------------------------";

        //print the array
        int numRows = (memPtr-PROGADDR+48)/16;
        memPtr = PROGADDR-16;
        memPtr -= (PROGADDR % 16);
        output += "\r\n\r\nAddress      Contents\r\n-------------------------------\r\n";
        for(int i=0; i<numRows; i++){
            output += IntToHex((int)memPtr, 4) + "       ";
            for(int j=0; j<16; j++){
                if(!(j%4)) output += "  ";
                if(memMask[memPtr] == 0)
                    output += "xx";
                else if(memMask[memPtr] == 1)
                    output += "..";
                else if(memMask[memPtr] == 2)
                    output += IntToHex((int)MEM[memPtr], 2);
                memPtr++;
            }
            output += "\r\n";
        }

        //print program location
        output += "\r\n\r\nProgram loaded at address: " + IntToHex((int)PROGADDR, 4) + "\r\n";
        output += "First executable instruction at: " + IntToHex((int)EXECADDR, 4) + "\r\n\r\n";

        String fileName = "outfile";
        printf("Writing output to \"outfile\"...\n");
        if(FileExists(fileName))
            DeleteFile(fileName);
        int iFileHandle = FileCreate(fileName);
        if(iFileHandle == -1)
            throw runtime_error("Could not create output file");
        FileWrite(iFileHandle, output.c_str(), output.Length());
        FileClose(iFileHandle);
    }__finally{
        delete [] MEM;
        delete [] memMask;
    }
}
//---------------------------------------------------------------------------


